# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines quadratic loss functions."""

import jax
import jax.numpy as jnp

HIGHEST = jax.lax.Precision.HIGHEST


def quadform(hess, x, precision):
  """Computes a quadratic form (x^T @ H @ x)."""
  u = jnp.dot(hess, x, precision=precision)  # u = Hx
  return jnp.inner(x, u, precision=precision)


def loguniform(n, lambda_min, lambda_max, precision=HIGHEST):
  """Quadratic loss function with loguniform eigenvalues.

  The loss is: f(x) = (1/2) x^T H x + x^T v + b.

  The eigenvalues of the Hessian (H) are sampled uniformly on a
  logarithmic grid from lambda_min to lambda_max.

  Args:
    n: int, Problem dimension (number of parameters).
    lambda_min: float, Minimum eigenvalue of the Hessian.
    lambda_max: float, Maximum eigenvalue of the Hessian.
    precision: Which lax precision to use (default: HIGHEST).

  Returns:
    problem_fun: Function that takes a jax PRNGkey and a precision argument
      and returns an (initial_params, loss_fun) tuple.
  """

  def problem_fun(key):
    """Builds a quadratic loss problem."""
    pkey, ekey, qkey, vkey = jax.random.split(key, 4)

    # Sample eigenvalues.
    log_eigenvalues = jax.random.uniform(ekey,
                                         shape=(n,),
                                         minval=lambda_min,
                                         maxval=lambda_max)
    eigenvalues = 10**log_eigenvalues

    # Build orthonormal basis.
    basis = jax.nn.initializers.orthogonal()(qkey, shape=(n, n))

    # Define hessian.
    hess = jnp.dot(jnp.dot(basis, jnp.diag(eigenvalues), precision=precision),
                   basis.T,
                   precision=precision)

    # Random vector for the linear term in the loss.
    v = jax.random.normal(vkey, shape=(n,))

    # Compute an offset such that the global minimum has a loss of zero.
    xstar = jnp.linalg.solve(hess, -v)
    offset = -0.5 * quadform(hess, xstar, precision=precision) - jnp.inner(v, xstar, precision=precision)  # pylint: disable=line-too-long

    def loss_fun(x, _):
      return 0.5 * quadform(hess, x, precision=precision) + jnp.inner(v, x, precision=precision) + offset  # pylint: disable=line-too-long

    x_init = jax.random.normal(pkey, shape=(n,))
    return x_init, loss_fun

  return problem_fun



"""Datasets."""

import tensorflow_datasets as tfds
import tensorflow_text as text
import tensorflow as tf

import os

from renn import utils
from renn.data.tokenizers import load_tokenizer, SEP
from renn.data import data_utils

__all__ = ['ag_news', 'goemotions', 'imdb', 'snli', 'tokenize_fun', 'mnist']


def pipeline(dset, preprocess_fun=utils.identity, filter_fn=None, bufsize=1024):
  """Common (standard) dataset pipeline.
  Preprocesses the data, filters it (if a filter function is specified), caches it, and shuffles it.

  Note: Does not batch"""

  # Apply custom preprocessing.
  dset = dset.map(preprocess_fun)

  # Apply custom filter.
  if filter_fn is not None:
    dset = dset.filter(filter_fn)

  # Cache and shuffle.
  dset = dset.cache().shuffle(buffer_size=bufsize)

  return dset


def tokenize_fun(tokenizer):
  """Standard text processing function."""
  wsp = text.WhitespaceTokenizer()
  return utils.compose(tokenizer.tokenize, wsp.tokenize, text.case_fold_utf8)


def padded_batch(dset, batch_size, sequence_length, label_shape=()):
  """Pads examples to a fixed length, and collects them into batches."""

  # We assume the dataset contains inputs, labels, and an index.
  padded_shapes = {
      'inputs': (sequence_length,),
      'labels': label_shape,
      'index': (),
  }

  # Filter out examples longer than sequence length.
  dset = dset.filter(lambda d: d['index'] <= sequence_length)

  # Pad remaining examples to the sequence length.
  dset = dset.padded_batch(batch_size, padded_shapes)

  return dset


def load_tfds(name, split, preprocess_fun, filter_fn=None, data_dir=None):
  """Helper that loads a text classification dataset
  from tensorflow_datasets"""

  # Load raw dataset.
  dset = tfds.load(name, split=split, data_dir=data_dir)

  # Apply common dataset pipeline.
  dset = pipeline(dset, preprocess_fun=preprocess_fun, filter_fn=filter_fn)

  return dset


def load_csv(name, split, preprocess_fun, filter_fn=None, data_dir=None):
  """Helper that loads a text classification dataset
  from a csv file"""
  # Load raw dataset.

  output_types = {'text': tf.string, 'label': tf.int64}
  output_shapes = {'text': (), 'label': ()}

  filename = os.path.join(data_dir, f'{split}.csv')
  row_parser = data_utils.PARSERS[name]
  dset_iterator = lambda: data_utils.readfile(filename, row_parser)
  dset = tf.data.Dataset.from_generator(dset_iterator, output_types,
                                        output_shapes)

  # Apply common dataset pipeline.
  dset = pipeline(dset, preprocess_fun=preprocess_fun, filter_fn=filter_fn)

  return dset


def ag_news(split,
            vocab_file,
            sequence_length=100,
            batch_size=64,
            transform_fn=utils.identity,
            filter_fn=None,
            data_dir=None):
  """Loads the ag news dataset."""
  tokenize = tokenize_fun(load_tokenizer(vocab_file))

  def _preprocess(d):
    """Applies tokenization."""
    tokens = tokenize(d['description']).flat_values  # Note: we ignore 'title'
    preprocessed = {
        'inputs': tokens,
        'labels': d['label'],
        'index': tf.size(tokens),
    }
    return transform_fn(preprocessed)

  # Load dataset.
  dset = load_tfds('ag_news_subset',
                   split,
                   _preprocess,
                   filter_fn,
                   data_dir=data_dir)

  # Pad remaining examples to the sequence length.
  dset = padded_batch(dset, batch_size, sequence_length)

  return dset


def goemotions(split,
               vocab_file,
               sequence_length=50,
               batch_size=64,
               emotions=None,
               transform=utils.identity,
               filter_fn=None,
               data_dir=None):
  """Loads the goemotions dataset."""
  tokenize = tokenize_fun(load_tokenizer(vocab_file))

  if emotions is not None:  # Use all emotions.
    emotions = ('admiration', 'amusement', 'anger', 'annoyance', 'approval',
                'caring', 'confusion', 'curiosity', 'desire', 'disappointment',
                'disapproval', 'disgust', 'embarrassment', 'excitement', 'fear',
                'gratitude', 'grief', 'joy', 'love', 'nervousness', 'neutral',
                'optimism', 'pride', 'realization', 'relief', 'remorse',
                'sadness', 'surprise')

  def _preprocess(d):
    tokens = tokenize(d['comment_text']).flat_values
    index = tf.size(tokens)
    labels = tf.convert_to_tensor([d[e] for e in emotions], dtype=tf.int64)
    preprocessed = {
        'inputs': tokens,
        'labels': labels,
        'index': index,
    }
    return transform(preprocessed)

  # Load dataset.
  dset = load_tfds('goemotions',
                   split,
                   _preprocess,
                   filter_fn,
                   data_dir=data_dir)

  # Pad remaining examples to the sequence length.
  dset = padded_batch(dset,
                      batch_size,
                      sequence_length,
                      label_shape=(len(emotions),))

  return dset


def imdb(split,
         vocab_file,
         sequence_length=1000,
         batch_size=64,
         transform=utils.identity,
         filter_fn=None,
         data_dir=None):
  """Loads the imdb reviews dataset."""
  tokenize = tokenize_fun(load_tokenizer(vocab_file))

  def _preprocess(d):
    """Applies tokenization."""
    tokens = tokenize(d['text']).flat_values
    preprocessed = {
        'inputs': tokens,
        'labels': d['label'],
        'index': tf.size(tokens),
    }
    return transform(preprocessed)

  # Load dataset.
  dset = load_tfds('imdb_reviews',
                   split,
                   _preprocess,
                   filter_fn,
                   data_dir=data_dir)

  # Pad remaining examples to the sequence length.
  dset = padded_batch(dset, batch_size, sequence_length)

  return dset


def yelp(split,
         num_classes,
         vocab_file,
         sequence_length=1000,
         batch_size=64,
         transform=utils.identity,
         filter_fn=None,
         data_dir=None):
  """Loads the yelp reviews dataset."""
  tokenize = tokenize_fun(load_tokenizer(vocab_file))

  if data_dir is None:
    raise ValueError('Yelp dataset requires data_dir to be provided.')

  def _preprocess(d):
    """Applies tokenization, and
    transforms the yelp labels according to the
    specified number of classes"""

    star_to_label = {
        1: {
            1: 0,
            2: 0,
            3: -1,
            4: 1,
            5: 1
        },
        2: {
            1: 0,
            2: 0,
            3: -1,
            4: 1,
            5: 1
        },
        3: {
            1: 0,
            2: -1,
            3: 1,
            4: -1,
            5: 2
        },
        5: {
            1: 0,
            2: 1,
            3: 2,
            4: 3,
            5: 4
        }
    }

    star_to_label = star_to_label[num_classes]
    tokens = tokenize(d['text']).flat_values
    preprocessed = {
        'inputs': tokens,
        'labels': star_to_label[d['label']],
        'index': tf.size(tokens),
    }

    return transform(preprocessed)

  filter_fn = lambda x: x['labels'] != -1

  # Load dataset.
  dset = load_csv('yelp', split, _preprocess, filter_fn, data_dir=data_dir)

  # Pad remaining examples to the sequence length.
  dset = padded_batch(dset, batch_size, sequence_length)

  return dset


def dbpedia(split,
            num_classes,
            vocab_file,
            sequence_length=1000,
            batch_size=64,
            transform=utils.identity,
            filter_fn=None,
            data_dir=None):
  """Loads the dpedia text classification dataset."""
  tokenize = tokenize_fun(load_tokenizer(vocab_file))

  if data_dir is None:
    raise ValueError('DBPedia dataset requires data_dir to be provided.')

  def _preprocess(d):
    """Applies tokenization, and
    transforms the dbpedia labels according to the
    specified number of classes

    For a given number of classes, the classes with
    labels below that number are kept, and all other classes
    are removed.

    So, e.g., num_classes = 4, would keep classes 0,1,2,3"""

    def relabel(label):
      if label <= num_classes:
        # in DBPedia csv file, labels are
        # given as 1, 2, ...
        return label - 1
      else:
        return tf.constant(-1, dtype=tf.int64)

    tokens = tokenize(d['text']).flat_values
    preprocessed = {
        'inputs': tokens,
        'labels': relabel(d['label']),
        'index': tf.size(tokens),
    }

    return transform(preprocessed)

  filter_fn = lambda x: x['labels'] != -1

  # Load dataset.
  dset = load_csv('dbpedia', split, _preprocess, filter_fn, data_dir=data_dir)

  # Pad remaining examples to the sequence length.
  dset = padded_batch(dset, batch_size, sequence_length)

  return dset


def snli(split,
         vocab_file,
         sequence_length=75,
         batch_size=64,
         transform=utils.identity,
         filter_fn=None,
         data_dir=None):
  """Loads the SNLI dataset."""
  tokenize = tokenize_fun(load_tokenizer(vocab_file))

  def _preprocess(d):
    """Applies tokenization."""
    hypothesis = tokenize(d['hypothesis']).flat_values
    premise = tokenize(d['premise']).flat_values
    sep = tokenize(SEP).flat_values
    tokens = tf.concat([hypothesis, sep, premise], axis=0)
    return transform({
        'inputs': tokens,
        'labels': d['label'],
        'index': tf.size(tokens),
    })

  # Load dataset.
  dset = load_tfds('snli', split, _preprocess, filter_fn, data_dir=data_dir)

  # Pad remaining examples to the sequence length.
  dset = padded_batch(dset, batch_size, sequence_length)

  return dset


def mnist(split,
          order='row',
          batch_size=64,
          transform=utils.identity,
          filter_fn=None,
          data_dir=None,
          classes=None):
  """Loads the serialized MNIST dataset.

  Args:
    classes - the subset of classes to keep.
              If None, all will be kept"""

  def _preprocess(example):
    image = tf.squeeze(example['image'])
    image = tf.cast(image, tf.float32) / 255.

    if order == 'col':
      image = tf.transpose(image, perm=[1, 0])

    return transform({'inputs': image, 'labels': example['label'], 'index': 28})

  # Load dataset.
  dset = tfds.load('mnist', data_dir=data_dir)[split]

  if classes is not None:
    # Filter out images without the proper label
    allowed_fn = _in_subset(classes)
    # Remap labels to be in range (0, number of classes)
    relabel_fn = _relabel_subset(classes)

    dset = dset.filter(allowed_fn).map(relabel_fn)

  dset = pipeline(dset, _preprocess, filter_fn)

  # Batch dataset.
  return dset.batch(batch_size)


def _relabel_subset(subclasses):
  """Provides a function for relabeling classes.
  Example should contain key 'label' """

  s = tf.constant(subclasses, dtype=tf.int64)

  def relabel(example):
    example.update({'label': tf.where(s == example['label'])[0][0]})
    return example

  return relabel


def _in_subset(subclasses):
  """Provides a function for determining whether
  an example is in one of the provided subclasses.
  Expmle should contain a key 'label' """

  s = tf.constant(subclasses, dtype=tf.int64)

  def in_subset(example):
    label = example['label']
    isallowed = tf.equal(s, label)
    reduced = tf.reduce_sum(tf.cast(isallowed, tf.float32))
    return tf.greater(reduced, tf.constant(0.))

  return in_subset



""" Data utils """
import csv


def readfile(filename, parse_row):
  """
  Reads a csv file containing labeled data, where
  the function parse_row() extracts a score and
  text from the labeled data
  """

  with open(filename, 'r') as f:
    for row in csv.reader(f):

      score, text = parse_row(row)

      yield {'text': text, 'label': score}


def column_parser(text_column):
  """
  Returns a parser which parses a row of a csv file
  containing labeled data, extracting the label
  and the text

  This parser assumes the label is the zeroth element
  of the row, and the text is the 'text_column' element
  """

  def f(row):
    return int(row[0]), row[text_column]

  return f


PARSERS = {'yelp': column_parser(1), 'dbpedia': column_parser(2)}



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Recurrent neural network (RNN) helper functions."""

import functools

import jax
from jax.experimental import stax
import jax.numpy as jnp
import numpy as np

from . import cells
from . import unroll

__all__ = ['build_rnn', 'mse', 'eigsorted', 'timescale']


def build_rnn(num_tokens, emb_size, cell, num_outputs=1):
  """Builds an end-to-end recurrent neural network (RNN) model.

  Args:
    num_tokens: int, Number of different input tokens.
    emb_size: int, Dimensionality of the embedding vectors.
    cell: RNNCell to use as the core update function (see cells.py).
    num_outputs: int, Number of outputs from the readout (Default: 1).

  Returns:
    init_fun: function that takes a PRNGkey and input_shape and returns
      expected shapes and initialized embedding, RNN, and readout parameters.
    apply_fun: function that takes a tuple of network parameters and batch of
      input tokens and applies the RNN to each sequence in the batch.
    emb_apply: function to just apply the embedding.
    readout_apply: function to just apply the readout.
  """
  emb_init, emb_apply = cells.embedding(num_tokens, emb_size)
  readout_init, readout_apply = stax.Dense(num_outputs)

  def init_fun(key, input_shape):
    """Initialize the components of the RNN.

    Args:
      key: Jax PRNGkey used to initialize the parameters.
      input_shape: tuple representing the input shape, should be
        (batch_size, sequence_length).

    Returns:
      shapes: set of tuples representing the shapes after applying
        the Embedding, RNN Cell, and Readout layers.
      network_params: tuple of network parameters, containing the
        embedding, RNN cell, and readout parameters.
    """

    emb_key, cell_key, readout_key = jax.random.split(key, 3)

    # Initialize the Embedding for the input tokens.
    emb_shape, emb_params = emb_init(emb_key, input_shape)

    # The cell is defined for a single update step, which is why we ignore
    # the sequence dimension (emb_shape[1]) here.
    rnn_shape, rnn_params = cell.init(cell_key, (emb_shape[0], emb_shape[2]))

    output_shape, readout_params = readout_init(readout_key, rnn_shape)

    shapes = (emb_shape, rnn_shape, output_shape)
    network_params = (emb_params, rnn_params, readout_params)

    return shapes, network_params

  def apply_fun(network_params, tokens):
    """Applies the RNN on a batch of input sequences.

    Args:
      network_params: tuple of network parameters (see init_fun).
      tokens: batch of inputs, with shape (batch_size, sequence_length).

    Returns:
      outputs: network outputs, at every step along the sequence.
    """
    emb_params, rnn_params, readout_params = network_params

    # Apply the embedding.
    inputs = emb_apply(emb_params, tokens)

    # Run the RNN.
    initial_states = cell.get_initial_state(rnn_params,
                                            batch_size=tokens.shape[0])
    return unroll.unroll_rnn(initial_states, inputs,
                             functools.partial(cell.batch_apply, rnn_params),
                             functools.partial(readout_apply, readout_params))

  return init_fun, apply_fun, emb_apply, readout_apply


def mse(y, yhat):
  """Mean squared error loss."""
  return 0.5 * jnp.mean((y - yhat)**2)


def eigsorted(jac):
  """Computes sorted eigenvalues and corresponding eigenvectors of a matrix.

  Notes:
    The eigenvectors are stored in the columns of the returned matrices.
    The right and left eigenvectors are returned, such that: J=REL^T

  Args:
    jac: numpy array used to compute the eigendecomposition (must be square).

  Returns:
    rights: right eigenvectors, as columns in the returned array.
    eigvals: numpy array of eigenvalues.
    lefts: left eigenvectors, as columns in the returned array.
  """
  unsorted_eigvals, unsorted_rights = np.linalg.eig(jac)
  sorted_indices = np.flipud(np.argsort(np.abs(unsorted_eigvals)))

  eigenvalues = unsorted_eigvals[sorted_indices]
  rights = unsorted_rights[:, sorted_indices]
  lefts = np.linalg.pinv(rights).T

  return rights, eigenvalues, lefts


def timescale(eigenvalues):
  """Converts eigenvalues into approximate time constants."""
  return -1. / np.log(np.abs(eigenvalues))



# coding=utf-8
# Copyright 2020 TF.Text Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Algorithm for learning wordpiece vocabulary."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import collections

Params = collections.namedtuple(
    'Params', 'upper_thresh lower_thresh '
    'num_iterations max_input_tokens '
    'max_token_length max_unique_chars vocab_size '
    'slack_ratio include_joiner_token joiner '
    'reserved_tokens')


def extract_char_tokens(word_counts):
  """Extracts all single-character tokens from word_counts.

  Args:
    word_counts: list of (string, int) tuples

  Returns:
    set of single-character strings contained within word_counts
  """

  seen_chars = set()
  for word, _ in word_counts:
    for char in word:
      seen_chars.add(char)
  return seen_chars


def ensure_all_tokens_exist(input_tokens, output_tokens, include_joiner_token,
                            joiner):
  """Adds all tokens in input_tokens to output_tokens if not already present.

  Args:
    input_tokens: set of strings (tokens) we want to include
    output_tokens: string to int dictionary mapping token to count
    include_joiner_token: bool whether to include joiner token
    joiner: string used to indicate suffixes

  Returns:
    string to int dictionary with all tokens in input_tokens included
  """

  for token in input_tokens:
    if token not in output_tokens:
      output_tokens[token] = 1

    if include_joiner_token:
      joined_token = joiner + token
      if joined_token not in output_tokens:
        output_tokens[joined_token] = 1

  return output_tokens


def get_split_indices(word, curr_tokens, include_joiner_token, joiner):
  """Gets indices for valid substrings of word, for iterations > 0.

  For iterations > 0, rather than considering every possible substring, we only
  want to consider starting points corresponding to the start of wordpieces in
  the current vocabulary.

  Args:
    word: string we want to split into substrings
    curr_tokens: string to int dict of tokens in vocab (from previous iteration)
    include_joiner_token: bool whether to include joiner token
    joiner: string used to indicate suffixes

  Returns:
    list of ints containing valid starting indices for word
  """

  indices = []
  start = 0
  while start < len(word):
    end = len(word)
    while end > start:
      subtoken = word[start:end]
      # Subtoken includes the joiner token.
      if include_joiner_token and start > 0:
        subtoken = joiner + subtoken
      # If subtoken is part of vocab, 'end' is a valid start index.
      if subtoken in curr_tokens:
        indices.append(end)
        break
      end -= 1

    if end == start:
      return None
    start = end

  return indices


def get_search_threshs(word_counts, upper_thresh, lower_thresh):
  """Clips the thresholds for binary search based on current word counts.

  The upper threshold parameter typically has a large default value that can
  result in many iterations of unnecessary search. Thus we clip the upper and
  lower bounds of search to the maximum and the minimum wordcount values.

  Args:
    word_counts: list of (string, int) tuples
    upper_thresh: int, upper threshold for binary search
    lower_thresh: int, lower threshold for binary search

  Returns:
    upper_search: int, clipped upper threshold for binary search
    lower_search: int, clipped lower threshold for binary search
  """

  counts = [count for _, count in word_counts]
  max_count = max(counts)
  min_count = min(counts)

  upper_search = max_count if max_count < upper_thresh else upper_thresh
  lower_search = min_count if min_count > lower_thresh else lower_thresh

  return upper_search, lower_search


def get_input_words(word_counts, reserved_tokens, max_token_length):
  """Filters out words that are longer than max_token_length or are reserved.

  Args:
    word_counts: list of (string, int) tuples
    reserved_tokens: list of strings
    max_token_length: int, maximum length of a token

  Returns:
    list of (string, int) tuples of filtered wordcounts
  """

  all_counts = []

  for word, count in word_counts:
    if len(word) > max_token_length or word in reserved_tokens:
      continue
    all_counts.append((word, count))

  return all_counts


def get_allowed_chars(all_counts, max_unique_chars):
  """Get the top max_unique_chars characters within our wordcounts.

  We want each character to be in the vocabulary so that we can keep splitting
  down to the character level if necessary. However, in order not to inflate
  our vocabulary with rare characters, we only keep the top max_unique_chars
  characters.

  Args:
    all_counts: list of (string, int) tuples
    max_unique_chars: int, maximum number of unique single-character tokens

  Returns:
    set of strings containing top max_unique_chars characters in all_counts
  """

  char_counts = collections.defaultdict(int)

  for word, count in all_counts:
    for char in word:
      char_counts[char] += count

  # Sort by count, then alphabetically.
  sorted_counts = sorted(sorted(char_counts.items(), key=lambda x: x[0]),
                         key=lambda x: x[1],
                         reverse=True)

  allowed_chars = set()
  for i in range(min(len(sorted_counts), max_unique_chars)):
    allowed_chars.add(sorted_counts[i][0])
  return allowed_chars


def filter_input_words(all_counts, allowed_chars, max_input_tokens):
  """Filters out words with unallowed chars and limits words to max_input_tokens.

  Args:
    all_counts: list of (string, int) tuples
    allowed_chars: list of single-character strings
    max_input_tokens: int, maximum number of tokens accepted as input

  Returns:
    list of (string, int) tuples of filtered wordcounts
  """

  filtered_counts = []
  for word, count in all_counts:
    if (max_input_tokens != -1 and len(filtered_counts) >= max_input_tokens):
      break
    has_unallowed_chars = False
    for char in word:
      if char not in allowed_chars:
        has_unallowed_chars = True
        break
    if has_unallowed_chars:
      continue
    filtered_counts.append((word, count))

  return filtered_counts


def generate_final_vocabulary(reserved_tokens, char_tokens, curr_tokens):
  """Generates final vocab given reserved, single-character, and current tokens.

  Args:
    reserved_tokens: list of strings (tokens) that must be included in vocab
    char_tokens: set of single-character strings
    curr_tokens: string to int dict mapping token to count

  Returns:
    list of strings representing final vocabulary
  """

  sorted_char_tokens = sorted(list(char_tokens))
  vocab_char_arrays = []
  vocab_char_arrays.extend(reserved_tokens)
  vocab_char_arrays.extend(sorted_char_tokens)

  # Sort by count, then alphabetically.
  sorted_tokens = sorted(sorted(curr_tokens.items(), key=lambda x: x[0]),
                         key=lambda x: x[1],
                         reverse=True)
  for token, _ in sorted_tokens:
    vocab_char_arrays.append(token)

  seen_tokens = set()
  # Adding unique tokens to list to maintain sorted order.
  vocab_words = []
  for word in vocab_char_arrays:
    if word in seen_tokens:
      continue
    seen_tokens.add(word)
    vocab_words.append(word)

  return vocab_words


def learn_with_thresh(word_counts, thresh, params):
  """Wordpiece learning algorithm to produce a vocab given frequency threshold.

  Args:
    word_counts: list of (string, int) tuples
    thresh: int, frequency threshold for a token to be included in the vocab
    params: Params namedtuple, parameters for learning

  Returns:
    list of strings, vocabulary generated for the given thresh
  """

  # Set of single-character tokens.
  char_tokens = extract_char_tokens(word_counts)
  curr_tokens = ensure_all_tokens_exist(char_tokens, {},
                                        params.include_joiner_token,
                                        params.joiner)

  for iteration in range(params.num_iterations):
    subtokens = [dict() for _ in range(params.max_token_length + 1)]
    # Populate array with counts of each subtoken.
    for word, count in word_counts:
      if iteration == 0:
        split_indices = range(1, len(word) + 1)
      else:
        split_indices = get_split_indices(word, curr_tokens,
                                          params.include_joiner_token,
                                          params.joiner)
        if not split_indices:
          continue

      start = 0
      for index in split_indices:
        for end in range(start + 1, len(word) + 1):
          subtoken = word[start:end]
          length = len(subtoken)
          if params.include_joiner_token and start > 0:
            subtoken = params.joiner + subtoken
          if subtoken in subtokens[length]:
            # Subtoken exists, increment count.
            subtokens[length][subtoken] += count
          else:
            # New subtoken, add to dict.
            subtokens[length][subtoken] = count
        start = index

    next_tokens = {}
    # Get all tokens that have a count above the threshold.
    for length in range(params.max_token_length, 0, -1):
      for token, count in subtokens[length].items():
        if count >= thresh:
          next_tokens[token] = count
        # Decrement the count of all prefixes.
        if len(token) > length:  # This token includes the joiner.
          joiner_len = len(params.joiner)
          for i in range(1 + joiner_len, length + joiner_len):
            prefix = token[0:i]
            if prefix in subtokens[i - joiner_len]:
              subtokens[i - joiner_len][prefix] -= count
        else:
          for i in range(1, length):
            prefix = token[0:i]
            if prefix in subtokens[i]:
              subtokens[i][prefix] -= count

    # Add back single-character tokens.
    curr_tokens = ensure_all_tokens_exist(char_tokens, next_tokens,
                                          params.include_joiner_token,
                                          params.joiner)

  vocab_words = generate_final_vocabulary(params.reserved_tokens, char_tokens,
                                          curr_tokens)

  return vocab_words


def learn_binary_search(word_counts, lower, upper, params):
  """Performs binary search to find wordcount frequency threshold.

  Given upper and lower bounds and a list of (word, count) tuples, performs
  binary search to find the threshold closest to producing a vocabulary
  of size vocab_size.

  Args:
    word_counts: list of (string, int) tuples
    lower: int, lower bound for binary search
    upper: int, upper bound for binary search
    params: Params namedtuple, parameters for learning

  Returns:
    list of strings, vocab that is closest to target vocab_size
  """

  thresh = (upper + lower) // 2
  current_vocab = learn_with_thresh(word_counts, thresh, params)
  current_vocab_size = len(current_vocab)

  # Allow count to be within k% of the target count, where k is slack ratio.
  slack_count = params.slack_ratio * params.vocab_size
  if slack_count < 0:
    slack_count = 0

  is_within_slack = (current_vocab_size <= params.vocab_size) and (
      params.vocab_size - current_vocab_size <= slack_count)

  # We've created a vocab within our goal range (or, ran out of search space).
  if is_within_slack or lower >= upper or thresh <= 1:
    return current_vocab

  current_vocab = None

  if current_vocab_size > params.vocab_size:
    return learn_binary_search(word_counts, thresh + 1, upper, params)

  else:
    return learn_binary_search(word_counts, lower, thresh - 1, params)


def learn(word_counts, params):
  """Takes in wordcounts and returns wordpiece vocabulary.

  Args:
    word_counts: list of (string, int) tuples
    params: Params namedtuple, parameters for learning

  Returns:
    string, final vocabulary with each word separated by newline
  """

  upper_search, lower_search = get_search_threshs(word_counts,
                                                  params.upper_thresh,
                                                  params.lower_thresh)
  all_counts = get_input_words(word_counts, params.reserved_tokens,
                               params.max_token_length)
  allowed_chars = get_allowed_chars(all_counts, params.max_unique_chars)
  filtered_counts = filter_input_words(all_counts, allowed_chars,
                                       params.max_input_tokens)

  vocab = learn_binary_search(filtered_counts, lower_search, upper_search,
                              params)

  return vocab



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Recurrent neural network (RNN) cells."""
import dataclasses

import jax
import jax.numpy as jnp
from jax.tree_util import register_pytree_node

__all__ = [
    'LinearRNN', 'RNNCell', 'StackedCell', 'GRU', 'LSTM', 'VanillaRNN', 'UGRNN',
    'embedding'
]

# Aliases for standard initializers and nonlinearities.
fan_in = jax.nn.initializers.variance_scaling(1., 'fan_in', 'normal')
zeros = jax.nn.initializers.zeros


@dataclasses.dataclass
class LinearRNN:
  """Dataclass for storing parameters of a Linear RNN."""
  A: jnp.array  # Input weights.      pylint: disable=invalid-name
  W: jnp.array  # Recurrent weights.  pylint: disable=invalid-name
  b: jnp.array  # Bias.

  def apply(self, x, h) -> jnp.array:
    """Linear RNN Update."""
    return self.A @ x + self.W @ h + self.b

  def flatten(self):
    return (self.A, self.W, self.b)


# Register the LinearRNN dataclass as a pytree, so that we can directly
# pass it to other jax functions (optimizers, flatten, etc.)
register_pytree_node(LinearRNN, lambda node: (node.flatten(), None),
                     lambda _, children: LinearRNN(*children))


class RNNCell:
  """Base class for all RNN Cells.

  An RNNCell must implement the following methods:
    init(PRNGKey, input_shape) -> output_shape, rnn_params
    apply(params, inputs, state) -> next_state
  """

  def __init__(self, num_units, h_init=zeros):
    """Initializes an RNNCell."""
    self.num_units = num_units
    self.h_init = h_init

    # Compute RNN Jacobians.
    self.inp_jac = jax.jacobian(self.apply, argnums=1)
    self.inp_jac.__doc__ = 'Computes the Jacobian of the RNN cell with respect to it\'s inputs.'  # pylint: disable=line-too-long

    self.rec_jac = jax.jacobian(self.apply, argnums=2)
    self.rec_jac.__doc__ = 'Computes the Jacobian of the RNN cell with respect to it\'s own hidden state.'  # pylint: disable=line-too-long

    # Generate cell update function that works on batches of inputs.
    self.batch_apply = jax.vmap(self.apply, in_axes=(None, 0, 0))
    self.batch_apply.__doc__ = 'Applies the RNN cell update to a batch of inputs and states.'  # pylint: disable=line-too-long

  def init(self, key, input_shape):
    raise NotImplementedError

  def apply(self, params, inputs, state):
    raise NotImplementedError

  def init_initial_state(self, key):
    return self.h_init(key, self.num_units)

  def get_initial_state(self, params, batch_size=None):
    """Gets initial RNN states.

    Args:
      params: rnn_parameters
      batch_size: batch size of initial states to create.

    Returns:
      An ndarray with shape (batch size, num_units).
    """
    if batch_size is None:
      return params['initial_state']
    else:
      initial_state = jnp.expand_dims(params['initial_state'], axis=0)
      return jnp.repeat(initial_state, batch_size, axis=0)


class StackedCell(RNNCell):
  """Stacks multiple RNN cells together.

  A stacked RNN cell is specified by a list of RNN cells and
  (optional) stax.Dense layers in between them.

  Note that the full hidden state for this cell is the concatenation
  of hidden states from all of the cells in the stack.
  """

  def __init__(self, layers):
    """Initializes a Stacked RNN Cell.

    Args:
      layers: list of layers, each of which is a (RNNCell, stax.Dense)
        tuple. The stax.Dense functions are used to transform the hidden
        state from the previous RNN in the stack to the inputs to the next
        RNN. If wish to feed the state in directly, use stax.Identity.
    """
    self.layers = layers

    # Total number of units across all cells in the stack.
    self.units = [c.num_units for c, _ in self.layers]
    super().__init__(sum(self.units))

  def init(self, key, input_shape):
    """Initializes parameters of a Stacked RNN."""

    shape = input_shape
    weights = []

    # Initialize each RNN cell and dense layer in the stack.
    for cell, (dense_init, _) in self.layers:
      key, cell_key, dense_key = jax.random.split(key, 3)
      cell_shape, cell_params = cell.init(cell_key, shape)
      shape, dense_params = dense_init(dense_key, cell_shape)
      weights.append((cell_params, dense_params))

    params = {
        'initial_state': super().init_initial_state(key),
        'weights': weights
    }

    return (input_shape[0], self.num_units), params

  def apply(self, params, inputs, state):
    """Applies a single step of a Stacked RNN."""

    # Split the RNN state vector into the state vector for each cell.
    states = jnp.split(state, jnp.cumsum(jnp.array(self.units))[:-1])
    next_states = []

    # Iteratively apply each RNN layer.
    for p, h, layer in zip(params['weights'], states, self.layers):

      # Get parameters and functions for this layer.
      cell_params, dense_params = p
      cell, (_, dense_apply) = layer

      # Apply the RNN and dense layers.
      next_state = cell.apply(cell_params, inputs, h)
      inputs = dense_apply(dense_params, h)

      # Store next state.
      next_states.append(next_state)

    return jnp.concatenate(next_states)


class VanillaRNN(RNNCell):
  """Vanilla RNN Cell."""

  def __init__(self, num_units, w_init=fan_in, b_init=zeros, h_init=zeros):
    """Vanilla RNN Cell.

    Args:
      num_units: int, Number of units in the RNN.
      w_init: Initializer for the recurrent & input weights (Default: fan_in).
      b_init: Initializer for the bias (Default: zeros).
      h_init: Initializer for the RNN initial condition (Default: zeros).
    """
    self.w_init = w_init
    self.b_init = b_init
    self.h_init = h_init
    super().__init__(num_units, h_init)

  def init(self, key, input_shape):
    """Initializes the parameters of a Vanilla RNN."""
    batch_size, num_inputs = input_shape

    rec_shape = (self.num_units, self.num_units)
    inp_shape = (self.num_units, num_inputs)
    bias_shape = (self.num_units,)
    output_shape = (batch_size, self.num_units)

    keys = jax.random.split(key, 4)
    weights = LinearRNN(self.w_init(keys[0], inp_shape),
                        self.w_init(keys[1], rec_shape),
                        self.b_init(keys[2], bias_shape))
    initial_state = super().init_initial_state(keys[3])
    params = {'initial_state': initial_state, 'weights': weights}
    return output_shape, params

  def apply(self, params, inputs, state):
    """Applies a single step of a Vanilla RNN."""
    linear_update = params['weights']
    return jnp.tanh(linear_update.apply(inputs, state))


class UGRNN(RNNCell):
  """Update-gate RNN Cell."""

  def __init__(self,
               num_units,
               gate_bias=0.,
               w_init=fan_in,
               b_init=zeros,
               h_init=zeros):
    """Update-gate RNN (UGRNN) Cell.

    Args:
      num_units: int, Number of units in the RNN.
      gate_bias: float, Bias to add to the update gate (Default: 0.).
      w_init: Initializer for the recurrent & input weights (Default: fan_in).
      b_init: Initializer for the bias (Default: zeros).
      h_init: Initializer for the RNN initial condition (Default: zeros).
    """
    self.gate_bias = gate_bias
    self.w_init = w_init
    self.b_init = b_init
    self.h_init = h_init
    super().__init__(num_units, h_init)

  def init(self, key, input_shape):
    batch_size, num_inputs = input_shape

    rec_shape = (self.num_units, self.num_units)
    inp_shape = (self.num_units, num_inputs)
    bias_shape = (self.num_units,)
    output_shape = (batch_size, self.num_units)

    def _build_linear_rnn(base_key):
      base_key, *keys = jax.random.split(base_key, 4)
      return base_key, LinearRNN(self.w_init(keys[0], inp_shape),
                                 self.w_init(keys[1], rec_shape),
                                 self.b_init(keys[2], bias_shape))

    # build internal gates / linear rnns.
    # we thread the `key` through the helper function `_build_linear_rnn`,
    # which returns a new key that we then pass to the next function call.
    key, update_gate = _build_linear_rnn(key)
    key, cell_state = _build_linear_rnn(key)

    key, key_ic = jax.random.split(key, 2)
    initial_state = super().init_initial_state(key_ic)
    params = {
        'initial_state': initial_state,
        'update_gate': update_gate,
        'cell_state': cell_state
    }
    return output_shape, params

  def apply(self, params, inputs, state):
    update_gate = params['update_gate']
    cell_state = params['cell_state']

    update = jax.nn.sigmoid(update_gate.apply(inputs, state) + self.gate_bias)
    cell = jnp.tanh(cell_state.apply(inputs, state))

    return update * state + (1 - update) * cell


class GRU(RNNCell):
  """Gated recurrent unit."""

  def __init__(self,
               num_units,
               gate_bias=0.,
               w_init=fan_in,
               b_init=zeros,
               h_init=zeros):
    """Gated recurrent unit (GRU) Cell.

    Args:
      num_units: int, Number of units in the RNN.
      gate_bias: float, Bias to add to the update gate (Default: 0.).
      w_init: Initializer for the recurrent & input weights (Default: fan_in).
      b_init: Initializer for the bias (Default: zeros).
      h_init: Initializer for the RNN initial condition (Default: zeros).
    """
    self.gate_bias = gate_bias
    self.w_init = w_init
    self.b_init = b_init
    self.h_init = h_init
    super().__init__(num_units, h_init)

  def init(self, key, input_shape):
    batch_size, num_inputs = input_shape

    rec_shape = (self.num_units, self.num_units)
    inp_shape = (self.num_units, num_inputs)
    bias_shape = (self.num_units,)
    output_shape = (batch_size, self.num_units)

    def _build_linear_rnn(base_key):
      base_key, *keys = jax.random.split(base_key, 4)
      return base_key, LinearRNN(self.w_init(keys[0], inp_shape),
                                 self.w_init(keys[1], rec_shape),
                                 self.b_init(keys[2], bias_shape))

    # build internal gates / linear rnns.
    # we thread the `key` through the helper function `_build_linear_rnn`,
    # which returns a new key that we then pass to the next function call.
    key, update_gate = _build_linear_rnn(key)
    key, reset_gate = _build_linear_rnn(key)
    key, cell_state = _build_linear_rnn(key)

    key, key_ic = jax.random.split(key, 2)
    initial_state = super().init_initial_state(key_ic)
    params = {
        'initial_state': initial_state,
        'update_gate': update_gate,
        'reset_gate': reset_gate,
        'cell_state': cell_state
    }
    return output_shape, params

  def apply(self, params, inputs, state):
    update_gate = params['update_gate']
    reset_gate = params['reset_gate']
    cell_state = params['cell_state']

    update = jax.nn.sigmoid(update_gate.apply(inputs, state) + self.gate_bias)
    reset = jax.nn.sigmoid(reset_gate.apply(inputs, state))
    cell = jnp.tanh(cell_state.apply(inputs, reset * state))

    return update * state + (1 - update) * cell


class LSTM(RNNCell):
  """Long-short term memory (LSTM)."""

  def __init__(self,
               num_units,
               forget_bias=1.,
               w_init=fan_in,
               b_init=zeros,
               h_init=zeros):
    """Long-short term memory (LSTM) Cell.

    Args:
      num_units: int, Number of units in the RNN.
      forget_bias: float, Bias to add to the forget gate. (Default: 1.0).
      w_init: Initializer for the recurrent & input weights (Default: fan_in).
      b_init: Initializer for the bias (Default: zeros).
      h_init: Initializer for the RNN initial condition (Default: zeros).
    """
    self.forget_bias = forget_bias
    self.w_init = w_init
    self.b_init = b_init
    self.h_init = h_init

    # Note that we double the number of units here, since the LSTM
    # contains units for both the hidden and cell states.
    # These are h(t) and c(t) in most LSTM equations.
    super().__init__(num_units * 2, h_init)

  def init(self, key, input_shape):
    batch_size, num_inputs = input_shape

    # We divide by 2 since for the LSTM, there are separate
    # updates for the cell and hidden state. The full state
    # will be split in two in the apply() method.
    n = int(self.num_units / 2)
    rec_shape = (n, n)
    inp_shape = (n, num_inputs)
    bias_shape = (n,)

    # The output contains the full number of units.
    output_shape = (batch_size, self.num_units)

    def _build_linear_rnn(key):
      key, *prngs = jax.random.split(key, 4)
      return key, LinearRNN(self.w_init(prngs[0], inp_shape),
                            self.w_init(prngs[1], rec_shape),
                            self.b_init(prngs[2], bias_shape))

    # Build internal gates / linear RNNs.
    # We thread the `key` through the helper function `_build_linear_rnn`,
    # which returns a new key that we then pass to the next function call.
    key, forget_gate = _build_linear_rnn(key)
    key, update_gate = _build_linear_rnn(key)
    key, output_gate = _build_linear_rnn(key)
    key, cell_state = _build_linear_rnn(key)

    key, key_ic = jax.random.split(key, 2)
    initial_state = super().init_initial_state(key_ic)
    params = {
        'initial_state': initial_state,
        'forget_gate': forget_gate,
        'update_gate': update_gate,
        'output_gate': output_gate,
        'cell_state': cell_state
    }
    return output_shape, params

  def apply(self, params, inputs, full_state):
    forget_gate = params['forget_gate']
    update_gate = params['update_gate']
    output_gate = params['output_gate']
    cell_state = params['cell_state']

    # Split full state into the hidden and cell states.
    state, cell = jnp.split(full_state, 2, axis=-1)

    f = jax.nn.sigmoid(forget_gate.apply(inputs, state) + self.forget_bias)
    i = jax.nn.sigmoid(update_gate.apply(inputs, state))
    o = jax.nn.sigmoid(output_gate.apply(inputs, state))
    c = jnp.tanh(cell_state.apply(inputs, state))

    new_cell = f * cell + i * c
    new_state = o * jnp.tanh(new_cell)

    return jnp.concatenate((new_state, new_cell), axis=-1)


def embedding(vocab_size,
              embedding_size,
              initializer=jax.nn.initializers.orthogonal()):
  """Builds a token embedding.

  Args:
    vocab_size: int, Size of the vocabulary (number of tokens).
    embedding_size: int, Dimensionality of the embedding.
    initializer: Initializer for the embedding (Default: orthogonal).

  Returns:
    init_fun: callable, Initializes the embedding given a key and input_shape.
    apply_fun: callable, Converts a set of tokens to embedded vectors.
  """

  def init_fun(key, input_shape):
    """Build an embedding matrix."""
    shape = (vocab_size, embedding_size)
    emb = initializer(key, shape)
    output_shape = input_shape + (embedding_size,)
    return output_shape, emb

  def apply_fun(emb, tokens, **kwargs):
    """Embedding lookup."""
    del kwargs  # unused
    return jnp.take(emb, tokens, axis=0)

  return init_fun, apply_fun



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for computing loss."""

import jax.numpy as jnp
from jax import lax

__all__ = ['binary_xent', 'multiclass_xent']


def binary_xent(logits, labels):
  """ Cross-entropy loss in in a two-class classification problem,
  where the model output is a single logit

  Args:
    logits: array of shape (batch_size, 1) or just (batch_size)
    labels: array of length batch_size, whose elements are either 0 or 1

  Returns:
    loss: scalar cross entropy loss
  """

  squeezed_logits = jnp.squeeze(logits)
  log_likelihood = jnp.maximum(squeezed_logits, 0) - squeezed_logits * labels + \
      jnp.log(1 + jnp.exp(-jnp.abs(squeezed_logits)))
  return jnp.mean(log_likelihood)


def multiclass_xent(logits, labels):
  # zero max of logit
  shifted = logits - lax.stop_gradient(logits.max(axis=-1, keepdims=True))
  log_probs = shifted - jnp.log(
      jnp.sum(jnp.exp(shifted), axis=-1, keepdims=True))

  log_likelihood = jnp.take_along_axis(log_probs,
                                       labels[:, jnp.newaxis],
                                       axis=1)
  xent_loss = -1 * jnp.mean(log_likelihood)

  return xent_loss



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for analysis."""

from collections import defaultdict
from itertools import product
import numpy as np

__all__ = ['pseudogrid']


def pseudogrid(coordinates, dimension):
  """Constructs a pseudogrid
    ('pseudo' in that it is not necessarily evenly-spaced)
    of points in 'dimension'-dimension space
    from the specified coordinates.

    Arguments:
    coordinates: a mapping between dimensions and
                  coordinates in those dimensions
    dimension: number of dimensions

    For all dimensions that are not specified, the coordinate
    is taken to be 0.

    Example:
      if coordinates = {0: [0, 1, 2],
                        2: [1]},
      and dimension = 4, the coordinates in dimensions
      1 and 3 will be taken as [0], yielding the effective
      coordinate-dictionary
        coordinates = {0: [0,1,2],
                       1: [0],
                       2: [1],
                       3: [0]}
      Then the resulting pseudogrid will be constructed as:
        [[0,0,1,0], [1,0,1,0], [2,0,1,0]]
  """

  all_coordinates = defaultdict(lambda: np.array(0.0))
  all_coordinates.update(coordinates)

  max_specified_dim = max(coordinates.keys())

  if max_specified_dim > 32:
    raise NotImplementedError('Maximum specified dimension cannot exceed 32')

  points = np.meshgrid(
      *[all_coordinates[i] for i in range(max_specified_dim + 1)])
  points = np.stack(points).reshape(max_specified_dim + 1, -1).T

  padding = np.zeros((points.shape[0], dimension - max_specified_dim - 1))

  return np.concatenate((points, padding), axis=1)



"""Update functions for common optimizers."""

import jax.numpy as jnp


def momentum(alpha, beta):

  def update(g, v):
    """Momentum update.

    Args:
      g: gradient
      v: velocity
    """
    v = beta * v + g
    return -alpha * v

  return update


def nesterov(alpha, beta):

  def update(g, v):
    """Nesterov momentum update.

    Args:
      g: gradient
      v: velocity
    """
    v = beta * v + g
    return -alpha * (beta * v + g)

  return update


def adagrad(alpha, beta):

  def update(g, g_sq, v):
    """Adagrad update.

    Args:
      g: gradient
      g_sq: cumulative squared gradient
      v: velocity
    """
    g_sq += jnp.square(g)
    g_norm = jnp.where(g_sq > 0, g / jnp.sqrt(g_sq), 0.)
    v = (1. - beta) * g_norm + beta * v
    return -alpha * v

  return update


def rmsprop(alpha, beta=0.9, eps=1e-5):

  def update(g, m):
    """RMSProp update.

    Args:
      g: gradient
      m: running average of the second moment
    """
    m = beta * m + jnp.square(g) * (1. - beta)
    g_norm = g / jnp.sqrt(m + eps)
    return -alpha * g_norm

  return update


def adam(alpha, beta1=0.9, beta2=0.999, eps=1e-5):

  def update(g, m, v):
    """Adam update.

    Note: this is uncorrected.

    Args:
      g: gradient
      v: running average of the first moment (momentum)
      m: running average of the second moment (normalization)
    """
    v = (1 - beta1) * g + beta1 * v  # First moment.
    m = (1 - beta2) * jnp.square(g) + beta2 * m  # Second moment.
    return -alpha * v / (jnp.sqrt(m) + eps)

  return update


def cwrnn(cell_apply, readout_apply):

  def update(g, h):
    """Component-wise RNN Optimizer update.

    Args:
      g: gradient
      h: RNN state
    """
    h_next = cell_apply(g, h)
    return readout_apply(h_next)

  return update



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for computing a scalar objective from a loss curve."""

import jax.numpy as jnp


def nanmin(fs):
  """Computes the NaN-aware minimum over the loss curve."""
  return jnp.nanmin(fs[1:]) / fs[0]


def final(fs):
  """Returns the final loss value."""
  return fs[-1] / fs[0]


def mean(fs):
  """Returns the average over the loss values."""
  return jnp.mean(fs) / fs[0]



"""Text processing."""

from collections import Counter
import itertools

from renn.data import wordpiece_tokenizer_learner_lib as vocab_learner

import tensorflow_text as text
import tensorflow as tf

__all__ = ['build_vocab', 'load_tokenizer']

# Special tokens
JOINER = '##'
UNK = '<unk>'
CLS = '<cls>'
SEP = '<sep>'


def build_vocab(corpus_generator, vocab_size, split_fun=str.split):
  """Builds a vocab file from a text generator."""

  # Split documents into words.
  words = itertools.chain(*map(split_fun, corpus_generator))

  # Count words in the corpus.
  word_counts = Counter(words)

  # Specify parameters.
  reserved_tokens = (UNK, CLS, SEP)
  params = vocab_learner.Params(upper_thresh=10000000,
                                lower_thresh=10,
                                num_iterations=4,
                                max_input_tokens=5000000,
                                max_token_length=50,
                                max_unique_chars=1000,
                                vocab_size=vocab_size,
                                slack_ratio=0.05,
                                include_joiner_token=True,
                                joiner=JOINER,
                                reserved_tokens=reserved_tokens)

  # Build the vocabulary.
  vocab = vocab_learner.learn(word_counts.items(), params)

  return vocab


def load_tokenizer(vocab_file, default_value=-1):
  """Loads a tokenizer from a vocab file."""

  # Build lookup table that maps subwords to ids.
  table = tf.lookup.TextFileInitializer(vocab_file, tf.string,
                                        tf.lookup.TextFileIndex.WHOLE_LINE,
                                        tf.int64,
                                        tf.lookup.TextFileIndex.LINE_NUMBER)
  static_table = tf.lookup.StaticHashTable(table, default_value)

  # Build tokenizer.
  tokenizer = text.WordpieceTokenizer(static_table,
                                      suffix_indicator=JOINER,
                                      max_bytes_per_word=100,
                                      max_chars_per_token=None,
                                      token_out_type=tf.int64,
                                      unknown_token=UNK)

  return tokenizer



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Recurrent neural network (RNN) cells."""

import jax
import jax.numpy as jnp

__all__ = ['unroll_rnn']


def identity(x):
  """Identity function f(x) = x."""
  return x


def unroll_rnn(initial_states, input_sequences, rnn_update, readout=identity):
  """Unrolls an RNN on a batch of input sequences.

  Given a batch of initial RNN states, and a batch of input sequences, this
  function unrolls application of the RNN along the sequence. The RNN state
  is updated using the `rnn_update` function, and the `readout` is used to
  convert the RNN state to outputs (defaults to the identity function).

  B: batch size.
  N: number of RNN units.
  T: sequence length.

  Args:
    initial_states: batch of initial states, with shape (B, N).
    input_sequences: batch of inputs, with shape (B, T, N).
    rnn_update: updates the RNN hidden state, given (inputs, current_states).
    readout: applies the readout, given current states. If this is the identity
      function, then no readout is applied (returns the hidden states).

  Returns:
    outputs: batch of outputs (batch_size, sequence_length, num_outputs).
  """

  def _step(state, inputs):
    next_state = rnn_update(inputs, state)
    outputs = readout(next_state)
    return next_state, outputs

  input_sequences = jnp.swapaxes(input_sequences, 0, 1)
  _, outputs = jax.lax.scan(_step, initial_states, input_sequences)

  return jnp.swapaxes(outputs, 0, 1)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Meta-optimization framework."""

from collections import defaultdict
import functools

import jax
import jax.numpy as jnp

import numpy as np

from .. import utils
from . import losses


def unroll_scan(initial_params, loss_fun, optimizer, num_steps, decorator):
  """Runs an optimizer on a given problem, using lax.scan.

  Note: this will cache parameters during the unrolled loop, and thus uses a
  lot of device memory, therefore it is not good for simply evaluating
  (testing) an optimizer. Instead, it is useful for when we need to compute
  a _derivative_ of some final loss with respect to the optimizer parameters.

  Args:
    initial_params: Initial parameters.
    loss_fun: A function that takes (params, step) and returns a loss.
    optimizer: A tuple containing an optimizer init function, an update
      function, and a get_params function.
    num_steps: int, number of steps to run the optimizer.
    decorator: callable, Optional decorator function used to wrap the
      apply_fun argument to lax.scan.

  Returns:
    final_params: Problem parameters after running the optimizer.
    fs: Loss at every step of the loop.
  """

  # Gradient of the loss function.
  f_df = jax.jit(jax.value_and_grad(loss_fun))

  # Get optimizer functions.
  opt_init, opt_update, get_params = optimizer

  # Build function that applies a single step of the optimizer.
  @decorator
  def _apply(state, step):
    """Applies one step of the optimizer."""
    params = get_params(state)  # Get inner parameters.
    f, df = f_df(params, step)  # Loss and gradient.
    next_state = opt_update(step, df, state)  # Step the optimizer.
    return next_state, f

  # Initialize and run the optimizer.
  initial_state = opt_init(initial_params)
  steps = jnp.arange(num_steps)
  final_state, fs = jax.lax.scan(_apply, initial_state, steps)

  return get_params(final_state), fs


def unroll_for(initial_params, loss_fun, optimizer, extract_state, steps):
  """Runs an optimizer on a given problem, using a for loop.

  Note: this is slower to compile than unroll_scan, but can be used to store
  intermediate computations (such as the optimizer state or problem
  parameters) at every iteration, for further analysis.

  Args:
    initial_params: Initial parameters.
    loss_fun: A function that takes (params, step) and returns a loss.
    optimizer: A tuple containing an optimizer init function, an update
      function, and a get_params function.
    extract_state: A function that given some optimizer state, returns
      what from that optimizer state to store. Note that each optimizer state
      is different, so this function depends on a particular optimizer.
    steps: A generator that yields integers from (0, num_steps).

  Returns:
    results: Dictionary containing results to save.
  """

  # Gradient of the loss function.
  f_df = jax.jit(jax.value_and_grad(loss_fun))

  # Get optimizer functions.
  opt_init, opt_update, get_params = optimizer
  opt_state = opt_init(initial_params)

  def extract(opt_state):
    """Function to extract state from a packed OptimizerState object."""
    states_flat, _, subtrees = opt_state
    full_states = map(jax.tree_unflatten, subtrees, states_flat)
    return list(map(extract_state, full_states))

  # Data structure to store intermediate computation.
  store = defaultdict(list)

  # Optimize
  for step in steps:

    # Query function to get loss and gradient.
    params = get_params(opt_state)  # Get parameters.
    loss, gradient = f_df(params, step)  # Loss and gradient.

    # Store current loss, parameters, and optimizer state.
    store['loss'].append(loss)
    store['params'].append(params)
    store['state'].append(extract(opt_state))
    store['gradient'].append(gradient)

    # Apply the optimizer.
    opt_state = opt_update(step, gradient, opt_state)

  # Collect results as numpy arrays.
  return {k: jax.device_get(v) for k, v in store.items()}


def build_metaobj(problem_fun,
                  optimizer_fun,
                  num_inner_steps,
                  meta_loss=losses.mean,
                  l2_penalty=0.0,
                  decorator=jax.remat):
  """Builds a meta-objective function.

  Args:
    problem_fun: callable, Takes a PRNGKey argument and returns initial
      parameters and a loss function.
    optimizer_fun: callable, Takes a PRNGKey argument and returns an
      optimizer tuple (as in jax.experimental.optimizers).
    num_inner_steps: int, Number of optimization steps.
    meta_loss: callable, Function to use to compute a scalar meta-loss.
    l2_penalty: float, L2 penalty to apply to the meta-parameters.
    decorator: callable, Optional function to wrap the apply_fun argument to
      lax.scan. By default, this is jax.remat, which will rematerialize the
      forward computation when computing the gradient, trading off computation
      for memory. Using the identity function will turn off remat.

  Returns:
    meta_objective: callable, Function that takes meta-parameters and a
      PRNGKey and returns a scalar meta-objective and the inner loss history.
  """

  def meta_objective(meta_params, prng_key):
    """Computes meta-loss for a given set of meta-parameters.

    Args:
      meta_params: Parameters of the optimizer.
      prng_key: Pseudorandom number generator key, used to sample the
        problem to train on.

    Returns:
      meta_objective_value: A scalar value that is the meta-objective.
      fs: An array of training losses (training curve).
    """

    # Build a problem.
    x0, loss_fun = problem_fun(prng_key)

    # Build an optimizer with these meta-parameters.
    opt = optimizer_fun(meta_params)

    # Run the optimizer for num_inner_steps.
    _, fs = unroll_scan(x0, loss_fun, opt, num_inner_steps, decorator)

    # Final objective with an l2 norm penalty.
    mobj = meta_loss(fs) + l2_penalty * utils.norm(meta_params)

    return mobj, fs

  return meta_objective


def evaluate(opt, problem_fun, num_steps, eval_key, num_repeats=64):
  """Evaluates an optimizer on a given problem.

  Args:
    opt: An optimizer tuple of functions (init_opt, update_opt, get_params)
      to evaluate.
    problem_fun: A function that returns an (initial_params, loss_fun,
      fetch_data) tuple given a PRNGKey.
    num_steps: Number of steps to run the optimizer for.
    eval_key: Base PRNGKey used for evaluation.
    num_repeats: Number of different evaluation seeds to use.

  Returns:
    losses: Array of loss values with shape (num_repeats, num_steps)
      containing the training loss curve for each random seed.
  """

  @jax.jit
  def _run(k):
    return unroll_scan(*problem_fun(k), opt, num_steps, utils.identity)[1]

  keys = jax.random.split(eval_key, num=num_repeats)
  return jax.device_get(jax.vmap(_run)(keys))


def outer_loop(key,
               initial_meta_params,
               meta_objective,
               meta_optimizer,
               steps,
               batch_size=1,
               save_every=None,
               clip_value=np.inf):
  """Meta-trains an optimizer.

  Args:
    key: Jax PRNG key, used for initializing the inner problem.
    initial_meta_params: pytree, Initial meta-parameters.
    meta_objective: function, Computes a (scalar) loss given meta-parameters
      and an array (batch) of random seeds.
    meta_optimizer: tuple of functions, Defines the meta-optimizer to use (for
      example, a jax.experimental.optimizers Optimizer tuple).
    steps: A generator that yields integers from (0, num_steps).
    batch_size: int, Number of problems to train per batch.
    save_every: int, Specifies how often to store auxiliary information. If
      None, then information is never stored (Default: None).
    clip_value: float, Specifies the gradient clipping value (maximum
      gradient norm) (Default: np.inf).

  Returns:
    final_params: Final optimized parameters.
    store: Dict containing saved auxiliary information during optimization.
  """
  # Store quantities during outer-optimization.
  store = defaultdict(list)

  # Build meta-optimizer.
  init_opt, update_opt, get_params = meta_optimizer
  mopt_state = init_opt(initial_meta_params)

  # Function to comppute the meta-gradient and meta-hessian.
  meta_val_and_grad = jax.value_and_grad(meta_objective)

  # Function to clip gradient values.
  clip_fun = functools.partial(clip, value=clip_value)

  @jax.jit
  def outer_step(key, step, state):
    """Single step of meta-optimization."""

    # Refresh random state.
    prng_key = jax.random.fold_in(key, step)
    prng_keys = jnp.stack(jax.random.split(prng_key, batch_size))

    # Get optimizer with the current meta-parameters.
    meta_params = get_params(state)

    # Evaluate the meta-objective and meta-gradient
    mobj, mgrad = meta_val_and_grad(meta_params, prng_keys)

    # Clip gradient values.
    clipped_mgrad = jax.tree_map(clip_fun, mgrad)

    # Update the optimizer
    state = update_opt(step, clipped_mgrad, state)

    return mobj, mgrad, state

  # Run outer optimization.
  for step in steps:
    mobj, mgrad, mopt_state = outer_step(key, step, mopt_state)

    # Optionally store information.
    if (save_every is not None) and (step % save_every) == 0:
      store['step'].append(step)
      store['mobj'].append(mobj)
      store['mgrad'].append(mgrad)

  final_params = get_params(mopt_state)
  store = {k: np.array(v) for k, v in store.items()}
  return final_params, store


def clip(x, value=jnp.inf):
  """Clips elements of x to have magnitude less than or equal to value."""

  # Guard to short circuit if no value is given.
  if value == jnp.inf:
    return x

  mask = (jnp.abs(x) <= value).astype(jnp.float32)
  return x * mask + value * (1. - mask) * jnp.sign(x)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serialization of pytrees."""

import enum
import numpy as np
import msgpack

from .rnn.cells import LinearRNN

__all__ = ['dump', 'load', 'dumps', 'loads']


def dump(pytree, file):
  return msgpack.pack(pytree, file, default=_ext_pack, strict_types=True)


def dumps(pytree):
  return msgpack.packb(pytree, default=_ext_pack, strict_types=True)


def load(file):
  return msgpack.unpack(file, ext_hook=_ext_unpack, raw=False)


def loads(bytes):
  return msgpack.unpackb(bytes, ext_hook=_ext_unpack, raw=False)


class _CustomExtType(enum.IntEnum):
  ndarray = 1
  tuple = 2
  linear_rnn = 3


def ndarray_to_bytes(arr):
  """Converts a numpy ndarray to bytes."""

  if arr.dtype.hasobject or arr.dtype.isalignedstruct:
    raise ValueError('Object and structured dtypes are not supported.')

  data = (arr.tobytes(), arr.dtype.str, arr.shape)
  return msgpack.packb(data, use_bin_type=True)


def bytes_to_ndarray(data):
  buffer, dtype, shape = msgpack.unpackb(data, raw=False)
  return np.frombuffer(buffer, dtype=dtype).reshape(shape)


def _ext_pack(x):
  if isinstance(x, np.ndarray):
    return msgpack.ExtType(_CustomExtType.ndarray, ndarray_to_bytes(x))

  elif isinstance(x, tuple):
    return msgpack.ExtType(_CustomExtType.tuple, dumps(list(x)))

  elif isinstance(x, LinearRNN):
    return msgpack.ExtType(_CustomExtType.linear_rnn, dumps(x.flatten()))

  return x


def _ext_unpack(code, data):
  if code == _CustomExtType.ndarray:
    return bytes_to_ndarray(data)

  elif code == _CustomExtType.tuple:
    return tuple(loads(data))

  elif code == _CustomExtType.linear_rnn:
    return LinearRNN(*loads(data))

  return msgpack.ExtType(code, data)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Load tasks from the library."""

import jax
import jax.numpy as jnp

from renn import utils
from sklearn.datasets import make_moons

from .task_lib import classification
from .task_lib import quadratic
from .task_lib import regression

__all__ = [
    'quad',
    'two_moons',
]

# Each task below is a function that takes problem parameters as
# arguments and returns a `problem_fun` function. This function takes a
# single argument, a PRNGKey, and returns a (x_init, loss_fun, data) tuple.
# x_init is a pytree initial problem parameters. loss_fun is a function
# that returns a scalar loss given parameters and a batch of data. Finally,
# data is an iterable pytree. All leaves of the tree must have the same first
# dimension, which is the number of steps to optimize for. These slices
# of data will be passed to the loss_fun during optimization.
quad = quadratic.loguniform


def two_moons(model, num_samples=1024, l2_pen=5e-3, seed=0):
  num_classes = 2
  x, y = make_moons(n_samples=num_samples,
                    shuffle=True,
                    noise=0.1,
                    random_state=seed)
  features = jnp.array(x)
  targets = jnp.array(y)

  return logistic_regression(model, features, targets, l2_pen=l2_pen)


def logistic_regression(model, features, targets, l2_pen=0.):
  """Helper function for logistic regression."""

  m, n = features.shape

  def problem_fun(prng_key):
    keys = jax.random.split(prng_key)
    input_shape = (-1, n)
    init_fun, predict_fun = model
    output_shape, x0 = init_fun(keys[0], input_shape)

    def loss_fun(x, step):
      del step
      logits = jnp.squeeze(predict_fun(x, features))
      data_loss = jnp.mean(jnp.log1p(jnp.exp(logits)) - targets * logits)
      reg_loss = l2_pen * utils.norm(x)
      return data_loss + reg_loss

    return x0, loss_fun

  return problem_fun



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Synthetic Datasets."""

import numpy as np
from itertools import product
from renn import utils

__all__ = ['Unordered']


def constant_sampler(value):
  """Returns a sampling function which always returns the value 'value'"""

  def sample(num_samples):
    return np.full((num_samples,), value)

  max_length = value

  return sample, max_length


def uniform_sampler(min_val, max_val):
  """returns a sampling function which samples uniformly between min_val and
  max_val, inclusive"""

  def sample(num_samples):
    return np.random.randint(min_val, max_val + 1, size=(num_samples,))

  max_length = max_val

  return sample, max_length


def build_vocab(valences=None, num_classes=3):
  """ Builds the vocabulary
  Vocabulary for this dataset consists of tuples, e.g., ('very', 3),
    indicating in this case a token which provides strong evidence of class 3.
  """

  if valences is None:
    valences = {
        'strongly_favor': 2,
        'favor': 1,
        'neutral': 0,
        'against': -1,
        'strongly_against': -2
    }

  words = product(valences, range(num_classes))

  def _score(word):
    """Converts a word like ('very', 1) to a
    vector-valued score, in this case (0,2,0,...)"""
    score = np.zeros(num_classes)
    score[word[1]] = valences[word[0]]
    return score

  vocab = {i: _score(word) for i, word in enumerate(words)}
  return vocab


class Unordered:
  """Synthetic dataset representing un-ordered classes, to mimic e.g.
  text-classification datasets like AG News (unlike, say, star-prediction or
  sentiment analysis, which features ordered classes"""

  def __init__(self,
               num_classes=3,
               batch_size=64,
               length_sampler='Constant',
               sampler_params={'value': 40}):

    SAMPLERS = {'Constant': constant_sampler, 'Uniform': uniform_sampler}

    self.num_classes = num_classes
    self.batch_size = batch_size

    if length_sampler in SAMPLERS.keys():
      self.sampler, self.max_len = SAMPLERS[length_sampler](**sampler_params)
    else:
      raise ValueError(f'length_sampler must be one of {SAMPLERS.keys()}')

    self.vocab = build_vocab(num_classes=num_classes)

  def __iter__(self):
    return self

  def __next__(self):
    """ Samples and returns a batch.  As with the real datasets,
    batch is a dictionary containing 'inputs', 'labels', and 'index' keys.
    'index' specifies the length of the sequence """
    lengths = self.sampler(num_samples=self.batch_size)
    max_length = max(lengths)

    batch = {
        'inputs':
            np.random.randint(len(self.vocab),
                              size=(self.batch_size, max_length)),
        'index':
            lengths
    }

    batch['labels'] = self.label_batch(batch)

    return batch

  def label_batch(self, batch):
    """ Calculates class labels for a batch of sentences """
    zipped = zip(batch['inputs'], batch['index'])

    class_scores = np.array([self.score(s, l) for s, l in zipped])
    return np.argmax(class_scores, axis=1)

  def score(self, sentence, length):
    """ Calculates the score, i.e. the amount of accumulated
    evidence in the sentence, for each class"""
    return sum([self.vocab[word] for word in sentence[:length]])



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Neural networks classification loss on mnist."""

import jax
from jax.experimental import stax
import jax.numpy as jnp

import numpy as np
import tensorflow_datasets as tfds

from ... import utils


def load_mnist():
  """Download and normalize the MNIST dataset."""
  # Fetch full datasets for evaluation
  mnist_data = tfds.load(name='mnist', batch_size=-1)
  mnist_data = tfds.as_numpy(mnist_data)

  datasets = {}
  for key in ('train', 'test'):
    data = mnist_data[key]
    images, labels = data['image'], data['label']
    datasets[key] = (np.array(images), np.array(labels))

  return datasets


def fully_connected(num_classes, layers=(64, 64)):
  """Build a fully connected neural network."""
  stack = [stax.Flatten]

  # Concatenate fully connected layers.
  for num_units in layers:
    stack += [stax.Dense(num_units), stax.Relu]

  # Output layer.
  stack += [stax.Dense(num_classes), stax.LogSoftmax]

  return stax.serial(*stack)


def conv2d(num_classes, layers=((32, 5, 2), (16, 3, 2), (16, 3, 2))):
  """Builds a simple convolutional neural network."""
  stack = []

  # Concatenate convolutional layers.
  for num_units, kernel_size, stride in layers:
    stack += [
        stax.Conv(num_units, (kernel_size, kernel_size), (stride, stride),
                  padding='SAME'), stax.Relu
    ]

  # Output layer.
  stack += [stax.Flatten, stax.Dense(num_classes), stax.LogSoftmax]

  return stax.serial(*stack)


def classifier(model, features, targets, batch_size=32, num_classes=10):
  """Builds a classification task using a fully connected network.

  Args:
    model: tuple, A jax.experimental.stax function pair.
    features: array with shape (num_examples, *feature_dimensions)
    targets: array with shape (num_examples, num_classes)
    batch_size: int, Number of examples to use per batch (Default: 32).
    num_classes: int, Number of output classes.

  Returns:
    problem_fun: function that takes a PRNGKey and returns initial
      parameters and a loss function.
  """

  # Prepare data.
  input_shape = (-1,) + features.shape[1:]
  num_examples = features.shape[0]

  # Model.
  init_fun, predict_fun = model

  def problem_fun(prng_key):
    """Builds a problem given a PRNG key.

    Args:
      prng_key: jax.random PRNG key.

    Returns:
      initial_params: pytree containing initial parameters.
      loss_fun: function that takes (params, step) and returns a scalar loss.
    """
    params_key, data_key = jax.random.split(prng_key)

    # Generate initial parameters for this problem.
    _, initial_params = init_fun(params_key, input_shape)

    # Classification loss function.
    def loss_fun(params, step):
      """Computes the cross-entropy classification loss."""

      # Subselect a minibatch of data.
      key = jax.random.fold_in(data_key, step)
      indices = jax.random.permutation(key, jnp.arange(num_examples))
      batch_indices = indices[:batch_size]
      batch_features = features[batch_indices]
      batch_targets = targets[batch_indices]

      # Preprocess features and targets.
      batch_features = batch_features.astype(jnp.float32) / 255.
      batch_targets = utils.one_hot(batch_targets, num_classes)

      # Evaluate loss.
      outputs = predict_fun(params, batch_features)
      return -jnp.mean(jnp.sum(outputs * batch_targets, axis=1))

    return initial_params, loss_fun

  return problem_fun



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for optimization."""

import jax
from jax import flatten_util
import jax.numpy as jnp

import numpy as np
import tqdm

__all__ = [
    'batch_mean', 'norm', 'identity', 'fst', 'snd', 'optimize', 'one_hot',
    'compose'
]


def batch_mean(fun, in_axes):
  """Converts a function to a batched version (maps over multiple inputs).

  This takes a function that returns a scalar (such as a loss function) and
  returns a new function that maps the function over multiple arguments (such
  as over multiple random seeds) and returns the average of the results.

  It is useful for generating a batched version of a loss function, where the
  loss function has stochasticity that depends on a random seed argument.

  Args:
    fun: function, Function to batch.
    in_axes: tuple, Specifies the arguments to fun to batch over. For
      example, in_axes=(None, 0) would batch over the second argument.

  Returns:
    batch_fun: function, computes the average over a batch.
  """
  mapped_fun = jax.vmap(fun, in_axes=in_axes)

  def batch_fun(*args):
    return jnp.mean(mapped_fun(*args))

  return batch_fun


def norm(params, order=2):
  """Computes the (flattened) norm of a pytree."""
  return jnp.linalg.norm(flatten_util.ravel_pytree(params)[0], ord=order)


def identity(x):
  """Identity function."""
  return x


def fst(xs):
  """Returns the first element from a list."""
  return xs[0]


def snd(xs):
  """Returns the second element from a list."""
  return xs[1]


def compose(*funcs):
  """Returns a function that is the composition of multiple functions."""

  def wrapper(x):
    for func in reversed(funcs):
      x = func(x)
    return x

  return wrapper


def optimize(loss_fun, x0, optimizer, steps, stop_tol=-np.inf):
  """Run an optimizer on a given loss function.

  Args:
    loss_fun: Scalar loss function to optimize.
    x0: Initial parameters.
    optimizer: An tuple of optimizer functions (init_opt, update_opt,
      get_params) from a jax.experimental.optimizers instance.
    steps: Iterator over steps.
    stop_tol: Stop if the loss is below this value (Default: -np.inf).

  Returns:
    loss_hist: Array of losses during training.
    final_params: Optimized parameters.
  """

  # Initialize optimizer.
  init_opt, update_opt, get_params = optimizer
  opt_state = init_opt(x0)

  # Loss and gradient.
  value_and_grad = jax.value_and_grad(loss_fun)

  @jax.jit
  def step(k, state):
    params = get_params(state)
    loss, grads = value_and_grad(params)
    return loss, update_opt(k, grads, state)

  # Store loss history.
  loss_hist = []
  for k in steps:
    f, opt_state = step(k, opt_state)
    loss_hist.append(f)

    if f <= stop_tol:
      break

  # Extract final parameters.
  final_params = get_params(opt_state)

  return np.array(loss_hist), final_params


def one_hot(labels, num_classes, dtype=jnp.float32):
  """Creates a one-hot encoding of an array of labels.

  Args:
    labels: array of integers with shape (num_examples,).
    num_classes: int, Total number of classes.
    dtype: optional, jax datatype for the return array (Default: float32).

  Returns:
    one_hot_labels: array with shape (num_examples, num_classes).
  """
  return jnp.array(jnp.array(labels)[:, None] == jnp.arange(num_classes), dtype)


def select(sequences, indices):
  """Given an array of shape (number_of_sequences, sequence_length, element_dimension),
  and a 1D array specifying which indices of each sequence to select, return
  a (number_of_sequences, element_dimension)-shaped array with the selected elements.

  Args:
    sequences: array with shape (number_of_sequences, sequence_length, element_dimension)
    indices: 1D array with length number_of_sequence

  Returns:
    selected_elements: array with shape (number_of_sequences, element_dimension)
  """

  assert len(indices) == sequences.shape[0]

  # shape indices properly
  indices_shaped = indices[:, jnp.newaxis, jnp.newaxis]

  # select element
  selected_elements = jnp.take_along_axis(sequences, indices_shaped, axis=1)

  # remove sequence dimension
  selected_elements = jnp.squeeze(selected_elements, axis=1)

  return selected_elements


def make_loss_function(network_apply_fun, basic_loss_fun, regularization_fun):
  """ Given the network-function, the basic loss function, and
  a regularization function, return a loss function which maps a tuple of
  network parameters and a training batch to a loss value.

  Arguments:
    network_apply_fun - maps (network_params, batched_inputs) -> network_logits
    basic_loss_fun - maps (logits, batched_labels) -> scalar loss value
    regularization_fun - maps network_params -> scalar loss value

  Returns:
    total_loss_fun - maps (network_params, batch) -> scalar loss value
  """

  def total_loss_fun(params, batch):
    """
    Maps network parameters and training batch to a loss value.

    Args:
      batch: a dictionary with keys ['inputs', 'index', 'labels']
        'inputs': sequence of inputs with shape (batch_size, max_sequence_length)
        'index' : 1d-array storing length of the corresponding input sequence
        'labels': 1d-array storing label of corresponding input sequence

    Returns:
      loss: scalar loss averaged over batch
    """

    all_time_logits = network_apply_fun(params, batch['inputs'])
    end_logits = select(all_time_logits, batch['index'] - 1)

    return basic_loss_fun(end_logits,
                          batch['labels']) + regularization_fun(params)

  return total_loss_fun


def make_acc_fun(network_apply_fun, num_outputs=1):
  """ Given a network function and number of outputs, returns an accuracy
  function """

  if num_outputs == 1:
    prediction_function = lambda x: (x >= 0.).astype(jnp.int32)
  else:
    prediction_function = lambda x: x.argmax(axis=-1).astype(jnp.int32)

  @jax.jit
  def accuracy_fun(params, batch):
    all_time_logits = network_apply_fun(params, batch['inputs'])
    end_logits = select(all_time_logits, batch['index'] - 1)
    predictions = jnp.squeeze(prediction_function(end_logits))
    accuracies = (batch['labels'] == predictions).astype(jnp.int32)
    return jnp.mean(accuracies)

  return accuracy_fun



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fixed point finding routines."""

import jax
import jax.numpy as jnp
import numpy as np
from scipy.spatial import distance

from .. import utils

__all__ = ['build_fixed_point_loss', 'find_fixed_points', 'exclude_outliers']


def build_fixed_point_loss(rnn_cell, cell_params):
  """Builds function to compute speed of hidden states.

  Args:
    rnn_cell: an RNNCell instance.
    cell_params: RNN parameters to use when applying the RNN.

  Returns:
    fixed_point_loss_fun: function that takes a batch of hidden states
      and inputs and computes the speed of the corresponding hidden
      states.
  """

  def fixed_point_loss_fun(h, x):
    """Computes the speed of hidden states.

    The speed is defined as the squared l2 distance between
    the current state and the next state, in response to a given
    input:

      Q = (1/2) || h - F(h, x) ||_2^2

    Args:
      h: The current state as a vector.
      x: The current input as a vector.

    Returns:
      fixed_point_loss_fun: A function that computes the fixed point speeds
        for a list or array of states.
    """
    h_next = rnn_cell.batch_apply(cell_params, x, h)
    return 0.5 * jnp.sum((h - h_next)**2, axis=1)

  return fixed_point_loss_fun


def find_fixed_points(fp_loss_fun,
                      initial_states,
                      x_star,
                      optimizer,
                      tolerance,
                      steps=range(1000)):
  """Run fixed point optimization.

  Args:
    fp_loss_fun: Function that computes fixed point speeds.
    initial_states: Initial state seeds.
    x_star: Input at which to compute fixed points.
    optimizer: A jax.experimental.optimizers tuple.
    tolerance: Stopping tolerance threshold.
    steps: Iterator over steps.

  Returns:
    fixed_points: Array of fixed points for each tolerance.
    loss_hist: Array containing fixed point loss curve.
    squared_speeds: Array containing the squared speed of each fixed point.
  """
  loss_hist, fps = utils.optimize(lambda h: jnp.mean(fp_loss_fun(h, x_star)),
                                  initial_states,
                                  optimizer,
                                  steps,
                                  stop_tol=tolerance)

  fixed_points = jax.device_get(fps)
  squared_speeds = jax.device_get(fp_loss_fun(fps, x_star))

  return fixed_points, loss_hist, squared_speeds


def exclude_outliers(points, threshold=np.inf, verbose=False):
  """Remove points that are not within some threshold of another point."""

  # Return all fixed points if tolerance is <= 0
  if np.isinf(threshold):
    return points

  # Return if there are less than two fixed points.
  if points.shape[0] <= 1:
    return points

  # Compute pairwise distances between all fixed points.
  distances = distance.squareform(distance.pdist(points))

  # Find distance to each nearest neighbor.
  nn_distance = np.partition(distances, 1, axis=0)[1]

  # Keep points whose nearest neighbor is within some distance threshold.
  keep_indices = np.where(nn_distance <= threshold)[0]

  # Log how many points were kept.
  if verbose:
    print(f'Keeping {len(keep_indices)} out of {len(points)} points.')

  return points[keep_indices]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from setuptools import setup, find_packages

global __version__
__version__ = None

with open('renn/version.py') as f:
  exec(f.read(), globals())

here = os.path.abspath(os.path.dirname(__file__))
try:
  README = open(os.path.join(here, 'README.md')).read()
except IOError:
  README = ''

setup(
    name='renn',
    version=__version__,
    description='Research tools for Reverse Engineering Neural Networks (RENN).',
    long_description=README,
    author='Niru Maheswaranathan',
    author_email='nirum@google.com',
    packages=find_packages(exclude=["examples"]),
    python_requires='>=3.7',
    install_requires=[
        'numpy >=1.12',
        'jax',
        'jaxlib',
        'msgpack',
        'sklearn',
        'tensorflow',
        'tensorflow-text',
        'tfds-nightly',
        'tqdm',
    ],
    url='https://github.com/google-research/reverse-engineering-neural-networks',
    license='Apache-2.0',
)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines regression tasks."""

import jax
import jax.numpy as jnp


def linear(dim, num_datapoints, scale, l2_penalty=1e-3):
  """Generates a random linear regression problem."""

  def problem_fun(prng_key):
    """Builds a problem given a PRNG key."""
    # Split the PRNG into two keys.
    params_key, features_key, targets_key = jax.random.split(prng_key, 3)

    # Generate a random feature matrix and target vector.
    features = scale * jax.random.normal(features_key, (num_datapoints, dim))
    targets = scale * jax.random.normal(targets_key, (num_datapoints,))

    # Generate initial parameters for this problem.
    initial_params = jax.random.normal(params_key, (dim,))

    # Quadratic loss function.
    def loss_fun(params, _):
      predictions = jnp.dot(features, params)
      loss = jnp.mean((predictions - targets)**2)
      return loss + l2_penalty * jnp.sum(params**2)

    return initial_params, loss_fun

  return problem_fun


def logistic(dim, num_datapoints, l2_penalty=1e-5):
  """Generates a random logistic regression problem."""

  def problem_fun(prng_key):
    """Builds a problem given a PRNG key."""

    # Split the PRNG into two keys.
    params_key, data_key = jax.random.split(prng_key, 2)
    features, targets = make_classification(data_key, dim, num_datapoints)

    # Generate initial parameters for this problem.
    initial_params = {
        'weights': jax.random.normal(params_key, (dim,)),
        'bias': jnp.zeros(1),
    }

    def loss_fun(params, _):
      """Logistic loss function."""
      logits = jnp.dot(features, params['weights']) + params['bias']

      data_loss = sigmoid_xent(logits, targets)
      reg_loss = l2_penalty * jnp.sum(params['weights']**2)
      return data_loss + reg_loss

    return initial_params, loss_fun

  return problem_fun


def sigmoid_xent(logits, labels):
  """Sigmoid cross-entropy (used for logistic regression)."""
  return jnp.mean(jnp.log(1 + jnp.exp(-logits)) + logits * (1 - labels))


def logistic_acc(logits, labels):
  """Logistic regression accuracy."""
  predictions = logits > 0.
  return jnp.mean(predictions == labels)


def make_classification(prng_key,
                        num_features,
                        num_examples,
                        num_classes=2,
                        sep=2.):
  """Generates random classification problems."""

  num_examples_per_class = num_examples // num_classes
  features, labels = [], []

  for label in range(num_classes):

    # Class mean is on a vertex of the hypercube.
    # e.g. mu0 = [sep,   0, 0, 0]
    #      mu1 = [  0, sep, 0, 0]
    # and so on.
    mu = sep * jnp.eye(num_features)[:, label]

    # Sample data from a multivariate normal centered at the class mean.
    keys = jax.random.split(prng_key, 2)
    sqrt_sigma = jax.random.normal(keys[0], (num_features, num_features))
    samples = jax.random.normal(keys[1], (num_examples_per_class, num_features))
    features.append(jnp.atleast_2d(mu) + jnp.dot(samples, sqrt_sigma))
    labels.append(label * jnp.ones(num_examples_per_class))

  return jnp.vstack(features), jnp.hstack(labels)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Define simple learned optimizer models."""

import numbers

import jax
from jax.experimental import optimizers
from jax.experimental import stax
import jax.numpy as jnp

# Aliases for standard initializers and nonlinearities.
fan_in = jax.nn.initializers.variance_scaling(1., 'fan_in', 'normal')
zeros = jax.nn.initializers.zeros


def append_to_sequence(sequence, element):
  """Appends an element to a rolling sequence buffer.

  Args:
    sequence: a sequence of ndarrays, concatenated along the first dimension.
    element: an ndarray to add to the sequence.

  Returns:
    sequence: the updated sequence, with the first element removed, the rest
      of the elements shifted over, and the new element added.
  """
  return jnp.vstack((sequence[1:], element[jnp.newaxis, ...]))


def cwrnn(key, cell, input_scale='raw', output_scale=1e-3):
  """Component-wise RNN Optimizer.

  This optimizer applies an RNN to update the parameters of each problem
  variable independently (hence the name, component-wise). It follows the
  same approach as in previous work (Andrychowicz et al 2016, Wichrowska
  et al 2017) that distribute the parameters along the batch dimension
  of the RNN. This allows us to easily update each parameter in parallel.

  Args:
    key: Jax PRNG key to use for initializing parameters.
    cell: An RNNCell to use (see renn/rnn/cells.py)
    input_scale: str, Specifies how to scale gradient inputs to the RNN. If
      'raw', then the gradients are not scaled. If 'log1p', then the scale
      and the sign of the inputs are split into a length 2 vector,
      [log1p(abs(g)), sign(g)].
    output_scale: float, Constant used to multiply (rescale) the RNN output.

  Returns:
    meta_parameters: A tuple containing the RNN parameters and the readout
      parameters. The RNN parameters themselves are a namedtuple. The readout
      parameters are also a tuple containing weights and a bias.

    optimizer_fun: A function that takes a set of meta_parameters and
      initializes an optimizer tuple containing functions to initialize the
      optimizer state, update the optimizer state, and get parameters from
      the optimizer state.
  """
  # Input and output shapes.
  n_in = 2 if input_scale == 'log1p' else 1
  n_out = 1

  # Initialize the readout
  readout_init, readout_apply = stax.Dense(n_out, W_init=zeros, b_init=zeros)

  # Initialize parameters.
  rnn_key, readout_key = jax.random.split(key)
  rnn_shape, rnn_params = cell.init(rnn_key, (None, n_in))
  _, readout_params = readout_init(readout_key, rnn_shape)
  initial_theta = (rnn_params, readout_params)

  @jax.experimental.optimizers.optimizer
  def optimizer_fun(theta):
    """Builds a component-wise RNN optimizer."""
    rnn_params, readout_params = theta

    def init_state(x):
      n = jnp.ravel(x).size
      return (x, cell.get_initial_state(rnn_params, batch_size=n))

    def update_opt(_, grads, state):
      x, h = state

      grad_vec = jnp.reshape(grads, (-1, 1))

      # Inputs are scaled by a constant factor.
      if isinstance(input_scale, numbers.Number):
        inputs = input_scale * grad_vec

      # Inputs are raw (unmodified) gradients.
      elif input_scale == 'raw':
        inputs = grad_vec

      # Inputs are the log-scale and sign of the gradient.
      elif input_scale == 'log1p':
        scale = jnp.log1p(jnp.abs(grad_vec))
        sign = jnp.sign(grad_vec)
        inputs = jnp.hstack((scale, sign))

      else:
        raise ValueError(f'Invalid input scale {input_scale}.')

      h_next = cell.batch_apply(rnn_params, inputs, h)
      outputs = readout_apply(readout_params, h_next)
      x_next = x + output_scale * jnp.reshape(outputs, x.shape)
      return (x_next, h_next)

    def get_params(state):
      return state[0]

    return (init_state, update_opt, get_params)

  return initial_theta, optimizer_fun


def lds(key, num_units, h0_init=zeros, w_init=fan_in):
  """Linear dynamical system (LDS) optimizer."""
  hstar_key, rec_key, inp_key, readout_key = jax.random.split(key, 4)

  # Initialize linear dynamical system
  h0 = h0_init(hstar_key, (num_units,))
  rec_jac = w_init(rec_key, (num_units, num_units))
  inp_jac = w_init(inp_key, (num_units, num_units))

  # Initialize the readout
  readout_init, readout_apply = stax.Dense(1, W_init=zeros, b_init=zeros)
  _, readout_params = readout_init(readout_key, (None, num_units))

  initial_meta_params = (h0, rec_jac, inp_jac, readout_params)

  @optimizers.optimizer
  def optimizer_fun(h_star, g_star, h_init, rec_jac, inp_jac, readout_params):
    """Linear dynamical system optimizer.

    Args:
      h_star: The state around which to linearize.
      g_star: The input around which to linearize.
      h_init: The initial state.
      rec_jac: Defines the recurrent dynamics.
      inp_jac: Multiplies the input gradients.
      readout_params: Tuple of (weights, biases).

    Returns:
      init_state: Initialize the optimizer state.
      update_opt: Updates the optimizer state variables given the current
        step, gradients, and current state.
      get_params: Gets parameters from the optimizer state.
    """

    def init_state(x):
      batch_size = jnp.ravel(x).size
      h = jnp.ones((batch_size, 1)) * jnp.reshape(h_init, (1, -1))
      return (x, h)

    def update_opt(_, grads, state):
      x, h = state
      g = jnp.reshape(grads, (-1, 1))
      h_next = h_star + jnp.dot(h - h_star, rec_jac.T) + jnp.dot(g - g_star, inp_jac.T)  # pylint: disable=line-too-long
      outputs = readout_apply(readout_params, h_next)
      x_next = x + jnp.reshape(outputs, x.shape)
      return (x_next, h_next)

    def get_params(state):
      return state[0]

    return (init_state, update_opt, get_params)

  return initial_meta_params, optimizer_fun


def linear(key, tau, scale, base=0):
  """Optimizer that is a linear function of gradient history."""
  initial_meta_params = base + scale * jax.random.uniform(key, (tau,))

  @optimizers.optimizer
  def optimizer_fun(meta_params):
    """Builds a linear optimizer with the given meta_params."""

    def init_fun(params):
      """Initialize optimizer state."""
      grad_seq = jnp.zeros((tau,) + params.shape)
      return (params, grad_seq)

    def update_fun(step, grads, state):
      """Apply a step of the optimizer."""
      del step  # Unused.
      params, grad_seq = state
      grad_seq = append_to_sequence(grad_seq, grads)
      params -= jnp.tensordot(meta_params, grad_seq, axes=1)
      return (params, grad_seq)

    def get_params(state):
      """Get parameters from the optimizer."""
      return state[0]

    return (init_fun, update_fun, get_params)

  return initial_meta_params, optimizer_fun


def linear_dx(key, tau, scale_grad, scale_dx, base_grad=0, base_gram=0):
  """Optimizer that is a linear function of gradient and parameter history."""

  key0, key1 = jax.random.split(key, 2)
  initial_meta_params = (base_grad +
                         scale_grad * jax.random.uniform(key0,
                                                         (tau,)), base_gram +
                         scale_dx * jax.random.uniform(key1, (tau - 1,)))

  @optimizers.optimizer
  def optimizer_fun(meta_params):
    """Builds a linear_dx optimizer."""
    theta_grad, theta_dx = meta_params

    def init_fun(params):
      """Initialize optimizer state."""
      grad_seq = jnp.zeros((tau,) + params.shape)
      param_seq = jnp.zeros((tau,) + params.shape)
      return (params, grad_seq, param_seq)

    def update_fun(step, grads, state):
      """Apply a step of the optimizer."""
      del step  # Unused.
      params, grad_seq, param_seq = state
      grad_seq = append_to_sequence(grad_seq, grads)
      param_seq = append_to_sequence(param_seq, params)

      # Differences in parameters.
      # TODO(nirum): This recomputes differences at every iteration. Should
      # time this to ensure that the repeated jnp.diff call is not too slow.
      delta_params = jnp.diff(param_seq, axis=0)

      grad_term = jnp.tensordot(theta_grad, grad_seq, axes=1)
      dx_term = jnp.tensordot(theta_dx, delta_params, axes=1)
      params -= (grad_term + dx_term)

      return (params, grad_seq, param_seq)

    def get_params(state):
      return state[0]

    return init_fun, update_fun, get_params

  return initial_meta_params, optimizer_fun


def gradgram(key, tau, scale_grad, scale_gram, base_grad=0, base_gram=0):
  """Optimizer that is a function of gradient history and inner products."""

  # Initialize meta-parameters.
  key0, key1 = jax.random.split(key, 2)

  initial_meta_params = (base_grad +
                         scale_grad * jax.random.uniform(key0,
                                                         (tau,)), base_gram +
                         scale_gram * jax.random.uniform(key1, (tau,)))

  # Generalized inner product.
  innerprod = jax.jit(
      jax.vmap(jax.vmap(lambda x, y: -jnp.sum(x * y), in_axes=(0, None)),
               in_axes=(None, 0)))

  # Batched norm.
  norms = jax.jit(jax.vmap(jnp.linalg.norm, in_axes=0))

  @optimizers.optimizer
  def optimizer_fun(meta_params):
    """An optimizer that uses gradient-gradient correlations."""
    theta_grad, theta_gram = meta_params

    def init_fun(params):
      """Initialize the optimizer state."""
      grad_seq = jnp.zeros((tau,) + params.shape)
      return (params, grad_seq)

    def update_fun(step, grads, state):
      """Apply a step of the optimzier."""
      del step  # Unused.
      params, grad_seq = state

      # Update gradient history.
      grad_seq = append_to_sequence(grad_seq, grads)

      # Compute normalized gram matrix.
      gram = innerprod(grad_seq, grad_seq)
      grad_norm = norms(grad_seq)
      gram /= (jnp.outer(grad_norm, grad_norm) + 1e-6)

      # Compute update terms.
      attn_weights = jnp.dot(stax.softmax(gram, axis=0), theta_gram)
      attn_term = jnp.tensordot(attn_weights, grad_seq, axes=1)
      grad_term = jnp.tensordot(theta_grad, grad_seq, axes=1)
      params -= (grad_term + attn_term)

      return (params, grad_seq)

    def get_params(state):
      return state[0]

    return init_fun, update_fun, get_params

  return initial_meta_params, optimizer_fun


def momentum(key):
  """Wrapper for the momentum optimizer."""
  del key  # Unused.
  initial_learning_rate = 1e-3
  initial_mass = 0.8

  def optimizer_fun(optimizer_params):
    return optimizers.momentum(*optimizer_params)

  return (initial_learning_rate, initial_mass), optimizer_fun


def aggmo(key, num_terms):
  """Aggregated momentum (aggmo)."""
  initial_learning_rate = 0.0
  initial_masses = zeros(key, (num_terms,))
  initial_meta_params = (initial_learning_rate, initial_masses)

  @optimizers.optimizer
  def optimizer_fun(v0, alphas, betas):
    """Aggregated momentum optimizer.

    Defines an aggregated momentum optimizer (momentum with multiple
    timescales). Instead of a single learning rate and momentum mass,
    this optimizer includes `n` of them.

    Args:
      v0: Initial velocity, with shape (1,) or (n,). If it is a single
        number, this will be broadcast along each of the n modes.
      alphas: Learning rate hyperparameters with shape (n,).
      betas: Momentum hyperparameters with shape (n,).

    Returns:
      init_state: Initialize the optimizer state.
      update_opt: Updates the optimizer state variables given the current
        step, gradients, and current state.
      get_params: Gets parameters from the optimizer state.
    """
    alphas = jnp.reshape(alphas, (1, -1))
    betas = jnp.reshape(betas, (1, -1))

    def init_state(x):
      n = jnp.ravel(x).size
      v = jnp.ones((n, 1)) * jnp.reshape(v0, (1, -1))
      return (x, v)

    def update_opt(_, grads, state):
      x, v = state
      inputs = jnp.reshape(grads, (-1, 1))

      v_next = betas * v - alphas * inputs
      x_next = x + jnp.real(jnp.sum(v_next, axis=1))
      return (x_next, v_next)

    def get_params(state):
      return state[0]

    return (init_state, update_opt, get_params)

  return initial_meta_params, optimizer_fun
